#![allow(dead_code)]

mod engine;

use component_macro::Component;
use engine::Component;
use engine::QuwiQuwi;
use engine::System;
use std::rc::Rc;

#[derive(Component)]
struct Sprite {
    sprite: engine::Sprite,
}

macro_rules! tuplify {
    ($t:ty) => {
        $t
    };
    ($t:ty, $($ts:ty),+) => {
        $t, tuplify!($($ts),+)
    };
}

macro_rules! run_quwi {
    ($ctx:expr, $t:ty) => {
        engine::Quwi::<$t>::new().run($ctx)
    };
    ($ctx:expr, $t:ty, $($ts:ty),+) => {
        engine::Quwi::<($t, tuplify!($($ts),+))>::new().run($ctx)
    };
}

#[derive(Component)]
struct Position(f64, f64);

#[derive(Clone, Component)]
struct Velocity(f64, f64);

struct VelocitySystem;
impl System for VelocitySystem {
    fn on_update(&self, ctx: &mut engine::Context, delta: f64) -> Result<(), engine::Error> {
        for id in run_quwi!(ctx, Velocity, Position) {
            let vel = ctx.entity_component::<Velocity>(id).clone();
            let Position(x, y) = ctx.entity_component::<Position>(id);
            *x += vel.0 * delta;
            *y += vel.1 * delta;
        }
        Ok(())
    }
}

#[derive(Component)]
struct Gravity;

struct GravitySystem;
impl System for GravitySystem {
    fn on_update(&self, ctx: &mut engine::Context, delta: f64) -> Result<(), engine::Error> {
        for id in run_quwi!(ctx, Gravity, Velocity) {
            let Velocity(_, y) = ctx.entity_component::<Velocity>(id);
            *y = if *y < 800.0 { *y + 400.0 * delta } else { *y };
        }
        Ok(())
    }
}

#[derive(Component)]
struct Cloud;

struct CloudSystem;

impl System for CloudSystem {
    fn on_update(&self, ctx: &mut engine::Context, delta: f64) -> Result<(), engine::Error> {
        let cloud_amount = ctx.entities_with_component::<Cloud>().len();
        if cloud_amount < 1 {
            let cloud = ctx.load_sprite("textures/clouds.png").unwrap();
            ctx.spawn(vec![
                Box::new(Cloud),
                Box::new(Sprite { sprite: cloud }),
                Box::new(Position(-100.0, 150.0)),
                Box::new(Velocity(0.0, 0.0)),
            ]);
        }

        for id in run_quwi!(ctx, Cloud, Velocity) {
            let Velocity(x, _) = ctx.entity_component::<Velocity>(id);
            *x = if *x < 200.0 { *x + 200.0 * delta } else { *x };
        }

        for id in run_quwi!(ctx, Cloud, Velocity) {
            let Position(x, _) = ctx.entity_component::<Position>(id);
            if *x > 1400.0 {
                ctx.despawn(id);
            }
        }
        Ok(())
    }
}

struct SpriteRenderer;
impl System for SpriteRenderer {
    fn on_update(&self, ctx: &mut engine::Context, _delta: f64) -> Result<(), engine::Error> {
        for id in run_quwi!(ctx, Sprite, Position) {
            let &mut Position(x, y) = ctx.entity_component::<Position>(id);
            let sprite = ctx.entity_component::<Sprite>(id).sprite;

            ctx.draw_sprite(sprite, x as i32, y as i32)?;
        }
        Ok(())
    }
}

#[derive(Component)]
struct PlayerMovement;

struct PlayerMovementSystem;
impl System for PlayerMovementSystem {
    fn on_update(&self, ctx: &mut engine::Context, _delta: f64) -> Result<(), engine::Error> {
        for id in run_quwi!(ctx, PlayerMovement, Velocity) {
            let d_down = ctx.key_pressed(engine::Keycode::D);
            let a_down = ctx.key_pressed(engine::Keycode::A);
            let Velocity(x, _) = ctx.entity_component::<Velocity>(id);
            *x = if d_down && !a_down {
                400.0
            } else if !d_down && a_down {
                -400.0
            } else {
                0.0
            };
        }
        Ok(())
    }
}

fn main() {
    let mut game = engine::Game::new().unwrap();

    let mut context = game.context();
    context.add_system(Rc::new(VelocitySystem));
    context.add_system(Rc::new(SpriteRenderer));
    context.add_system(Rc::new(GravitySystem));
    context.add_system(Rc::new(PlayerMovementSystem));
    context.add_system(Rc::new(CloudSystem));
    let player = context.load_sprite("textures/player.png").unwrap();
    let background = context.load_sprite("textures/mountains.png").unwrap();

    context.spawn(vec![
        Box::new(Sprite { sprite: background }),
        Box::new(Position(0.0, 0.0)),
    ]);

    context.spawn(vec![
        Box::new(Sprite { sprite: player }),
        Box::new(Position(16.0, 500.0)),
        Box::new(Velocity(0.0, -600.0)),
        Box::new(Gravity),
        Box::new(PlayerMovement),
    ]);
    game.run();
}
